//=====================================================================
//
//  File:       Buffer.cs
//  Summary:    This file contains the implementation of several classes that support the
//              Distinct transformation. These class are used to 
//              cache the rows of the incoming input buffers from the upstream 
//              component.
//              It is heavily based on RemoveDuplicates sample from Microsoft, the code
//              ha been adapted and simplified to reach a very simple and memory safe
//              component in order to have the DISTINCT of some columns in a flow.    
//  Date:       07 february 2007
//  Web:        www.sqlbi.eu
//  Mail:       alberto.ferrari@sqlbi.eu
//
//===================================================================== 

using System;
using System.Diagnostics;
using System.Collections;
using Microsoft.SqlServer.Dts.Pipeline;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;
using Microsoft.SqlServer.Dts.Runtime.Wrapper;
using System.Globalization;
using System.Collections.Generic;

namespace SqlBi.SSIS.Components {
    /// <summary>
    /// 
    /// This class is used to cache the rows received by the RemoveDuplicates transformation during ProcessInput.
    /// 
    /// </summary>
    public class Buffer {
        //
        //  This class has been created in order to let differente implementation of
        //  the buffer handling easy as during development we have tried several 
        //  kind of different sorted lists.
        //
        //  At the end it seemed to us that Dictionary is the fastest one for the
        //  kind of operation we need to carry on.
        //
        public class BufferRows  {
            private Dictionary<Row, Object> Rows = new Dictionary<Row, Object> ();

            public bool Add (Row rowToAdd) {
                if (! Rows.ContainsKey (rowToAdd)) {
                    Rows.Add (rowToAdd, null);
                    return true;
                } else {
                    return false;
                }
            }
        };

        private BufferRows rows;                // A container for each row
        private ColumnInfos columnInfos;        // The remapper in and out from the pipeline buffer

        /// <summary>
        /// 
        /// Standard Buffer creator
        /// 
        /// </summary>
        /// <param name="columnInfos"></param>
        public Buffer (ColumnInfos columnInfos) {
            rows = new BufferRows ();
            this.columnInfos = columnInfos;
        }

        /// <summary>
        /// 
        /// Creates a new row and returns it. It has been done in this way because
        /// the component will need the newly inserted row in order to send it to
        /// the output buffer, if needed
        /// 
        /// </summary>
        /// <param name="buffer"></param>
        /// <returns></returns>
        public Row NewRow (PipelineBuffer buffer) {
            object[] cols = new object[this.columnInfos.Count];

            for (int col = 0; col < columnInfos.Count; col++) {
                cols[col] = buffer[columnInfos[col].InputBufferColumnIndex];
            }
            return (new Row (cols));
        }

        /// <summary>
        /// 
        /// Copy the columns from the input buffer into 
        /// the internal buffer. The InputBufferColumnIndex property contains the index of the 
        /// column in the input buffer.
        /// 
        /// Returns 
        /// 
        /// </summary>
        /// <param name="buffer">The PipelineBuffer provided to the component in ProcessInput.</param>
        /// <returns>True if the column is new, false otherwise</returns>
        public bool AddRow (Row row) {
            return (rows.Add (row));
        }
    }

    /// <summary>
    /// 
    /// Simple Array class containg rows and columns. The class implements the IComparable interface
    /// so the buffer class can sort the array of row objects that it contains.
    /// 
    /// The hash code is computed once upon creation so when successive calls to gather it will be done
    /// there will be no computational overhead.
    /// 
    /// </summary>
    public class Row : IComparable {
        private object[] columnData;    // Contains the values of each of the columns in the row.
        public int hashCode;            // the Hash Code for the row

        /// <summary>
        /// 
        /// Standard constructor
        /// 
        /// </summary>
        /// <param name="columnValues"></param>
        public Row (object[] columnValues) {
            columnData = columnValues;
            hashCode = 0;
            for (int i = 0; i < columnData.Length; i++) {
                if (columnData[i] != null) {
                    hashCode += columnData[i].GetHashCode ();
                }
            }
        }

        /// <summary>
        /// 
        /// Provides access to the individual columns in the row.
        /// 
        /// </summary>
        /// <param name="index">The location of the column to retreive.</param>
        /// <returns>The value stored in the column.</returns>
        public object this[int index] {
            get {
                return columnData[index];
            }
            set {
                columnData[index] = value;
            }
        }

        public override int GetHashCode () {
            return hashCode;
        }

        public override bool Equals (object obj) {
            Row r = obj as Row;
            return this.Equals (r);
        }


        public static bool operator == (Row rowA, Row rowB) {
            if (rowA == null)
                throw new ArgumentNullException ("rowA");

            return rowA.Equals (rowB);
        }

        public static bool operator != (Row rowA, Row rowB) {
            if (rowA == null) {
                throw new ArgumentNullException ("rowA");
            }

            return !rowA.Equals (rowB);
        }

        public static bool operator < (Row rowA, Row rowB) {
            throw new NotImplementedException (Properties.Resources.OperatorLT);
        }

        public static bool operator > (Row rowA, Row rowB) {
            throw new NotImplementedException (Properties.Resources.OperatorGT);
        }

        /// <summary>
        /// 
        /// Checks to see if the provided row is an exact duplicate of this row.
        /// 
        /// </summary>
        /// <param name="row">The row to compare.</param>
        /// <returns>true if each column in the row is exactly the same; otherwise false</returns>
        public bool Equals (Row row) {
            for (int x = 0; x < this.columnData.Length; x++) {
                if (columnData[x] == null && row.columnData[x] != null)
                    return false;
                else if (columnData[x] != null && row.columnData[x] == null)
                    return false;
                else if (columnData[x] != null & row.columnData[x] != null) {
                    if (CompareColumns (columnData[x], row.columnData[x]) != 0)
                        return false;
                }
            }
            return true;
        }

        /// <summary>
        /// 
        /// Compare two rows.
        /// 
        /// </summary>
        /// <param name="obj">A row to compare to this row.</param>
        /// <returns>
        /// Less than zero;  This instance is less than obj.  
        /// Zero; This instance is equal to obj.  
        /// Greater than zero;  This instance is greater than obj.  
        /// </returns>
        public int CompareTo (object obj) {
            Row compareRow = obj as Row;

            //
            // Compare the columns of each row.
            //
            for (int col = 0; col < columnData.Length; col++) {
                if (columnData[col] != null && compareRow.columnData[col] == null)
                    return 1;
                else if (columnData[col] == null && compareRow.columnData[col] != null)
                    return -1;
                else {
                    if (columnData[col] != null && compareRow.columnData[col] != null) {
                        int result = CompareColumns (columnData[col], compareRow[col]);

                        if (result != 0)
                            return result;
                    }
                }
            }
            return 0;
        
        }
        /// <summary>
        /// 
        /// Compares two objects for equality.
        /// 
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes")]
        private static int CompareColumns (object colA, object colB) {
            Type aType = colA.GetType ();
            Type bType = colB.GetType ();

            if (!aType.Equals (bType))
                throw new Exception (Properties.Resources.TypeMismatchColumns);
            if (aType.Equals (typeof (string)))
                return ((string) colA).CompareTo ((string) colB);
            else if (aType.Equals (typeof (DateTime)))
                return Convert.ToDateTime (colA, CultureInfo.CurrentUICulture).CompareTo (Convert.ToDateTime (colB, CultureInfo.CurrentUICulture));
            else if (aType.Equals (typeof (decimal)))
                return ((decimal) colA).CompareTo ((decimal) colB);
            else if (aType.Equals (typeof (Guid)))
                return ((Guid) colA).CompareTo ((Guid) colB);
            else if (aType.Equals (typeof (Int16)))
                return ((Int16) colA).CompareTo ((Int16) colB);
            else if (aType.Equals (typeof (Int32)))
                return ((Int32) colA).CompareTo ((Int32) colB);
            else if (aType.Equals (typeof (Int64)))
                return ((Int64) colA).CompareTo ((Int64) colB);
            else if (aType.Equals (typeof (bool)))
                return ((bool) colA).CompareTo ((bool) colB);
            else if (aType.Equals (typeof (float)))
                return ((float) colA).CompareTo ((float) colB);
            else if (aType.Equals (typeof (double)))
                return ((double) colA).CompareTo ((double) colB);
            else if (aType.Equals (typeof (byte)))
                return ((byte) colA).CompareTo ((byte) colB);
            else if (aType.Equals (typeof (UInt16)))
                return ((UInt16) colA).CompareTo ((UInt16) colB);
            else if (aType.Equals (typeof (UInt32)))
                return ((UInt32) colA).CompareTo ((UInt32) colB);
            else if (aType.Equals (typeof (UInt64)))
                return ((UInt64) colA).CompareTo ((UInt64) colB);
            else
                throw new Exception (Properties.Resources.UnsupportedColumnDataType);
        }
    }

    /// <summary>
    /// 
    /// A simple collection class, that provides access the underlying array of ColumnInfo objects.
    /// 
    /// </summary>
    public class ColumnInfos {
        ArrayList columnInfos;
        public ColumnInfos () {
            columnInfos = new ArrayList ();
        }
        public ColumnInfo this[int index] {
            get {
                return (ColumnInfo) columnInfos[index];
            }
            set {
                columnInfos[index] = value;
            }
        }
        public int Add (ColumnInfo columnInfo) {
            return columnInfos.Add (columnInfo);
        }
        public int Count {
            get {
                return columnInfos.Count;
            }
        }

        /// <summary>
        /// 
        /// Stores the column information of the provided input and output objects in a ColumnInfo object. 
        /// This method is called by the RemoveDuplicates component during PreExecute when the BufferManager 
        /// is available, and assumes the exact same number of input and output columns.
        /// 
        /// </summary>
        /// <param name="bufferManager">The IDTSBufferManager90; used to locate the columns 
        /// in the input and output buffers.</param>
        /// <param name="output">The output of the component.</param>
        /// <param name="input">The input of the component.</param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public void AddColumnInformation (IDTSBufferManager90 bufferManager, IDTSOutput90 output, IDTSInput90 input) {
            if (bufferManager == null) {
                throw new ArgumentNullException ("bufferManager");
            }

            if (input == null) {
                throw new ArgumentNullException ("input");
            }

            if (output == null) {
                throw new ArgumentNullException ("output");
            }

            if (input.InputColumnCollection.Count != output.OutputColumnCollection.Count) {
                throw new Exception (Properties.Resources.InputColumnCollectionNotMatchOutput);
            }

            for (int x = 0; x < output.OutputColumnCollection.Count; x++) {
                ColumnInfo colInfo = new ColumnInfo ();
                IDTSOutputColumn90 outCol = output.OutputColumnCollection[x];
                IDTSInputColumn90 inCol = input.InputColumnCollection[x];

                //    Set the buffer column index for the input column and the output column.
                colInfo.InputBufferColumnIndex = bufferManager.FindColumnByLineageID (input.Buffer, inCol.LineageID);
                colInfo.OutputBufferColumnIndex = bufferManager.FindColumnByLineageID (output.Buffer, outCol.LineageID);

                //    Save the column
                columnInfos.Add (colInfo);
            }
        }
    }

    /// <summary>
    /// 
    /// Represents a column in an output or input object.
    /// 
    /// </summary>
    public struct ColumnInfo {
        private int inputBufferColumnIndex;
        private int outputBufferColumnIndex;

        public override bool Equals (object obj) {
            return base.Equals (obj);
        }

        public override int GetHashCode () {
            return base.GetHashCode ();
        }

        public static bool operator == (ColumnInfo ciA, ColumnInfo ciB) {
            return ciA.Equals (ciB);
        }

        public static bool operator != (ColumnInfo ciA, ColumnInfo ciB) {
            return !ciA.Equals (ciB);
        }

        public int InputBufferColumnIndex {
            get {
                return inputBufferColumnIndex;
            }
            set {
                inputBufferColumnIndex = value;
            }
        }

        public int OutputBufferColumnIndex {
            get {
                return outputBufferColumnIndex;
            }
            set {
                outputBufferColumnIndex = value;
            }
        }
    }
}
