using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Windows.Forms;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;

namespace SqlBi.SSIS.Components {
    [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Performance", "CA1812:AvoidUninstantiatedInternalClasses")]
    class DistinctUI : DataFlowComponentUI {
        const string InputColumnLineageIdPropName = "InputColumnLineageId";

        /// <summary>
        /// 
        /// Communication with UI form goes through these events. The UI will raise events when some data/action
        /// is needed and this class is responsible to answer those requests. This way we create separation between UI 
        /// specific logic and interactions with SSIS data flow object model.
        /// 
        /// </summary>
        /// <param name="form"></param>
        private void HookupEvents (DistinctForm form) {
            form.GetAvailableColumns += new GetAvailableColumnsEventHandler (form_GetAvailableColumns);
            form.SetInputOutputColumns += new ChangeInputOutputColumnsEventHandler (form_SetInputOutputColumns);
            form.DeleteInputOutputColumns += new ChangeInputOutputColumnsEventHandler (form_DeleteInputOutputColumns);
        }

        /// <summary>
        /// 
        /// Implementation of the method resposible for displaying the form.
        /// This one is abstract in the base class.
        /// 
        /// </summary>
        /// <param name="parentControl"></param>
        /// <returns></returns>
        protected override bool EditImpl (IWin32Window parentControl) {
            using (DistinctForm form = new DistinctForm ()) {
                this.HookupEvents (form);

                if (form.ShowDialog (parentControl) == DialogResult.OK) {
                    return true;
                } else {
                    return false;
                }
            }
        }

        /// <summary>
        /// 
        /// This method is called by the Form in order to get the available columns
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        void form_GetAvailableColumns (object sender, AvailableColumnsArgs args) {
            Debug.Assert (this.VirtualInput != null, "Virtual input is not valid.");

            this.ClearErrors ();

            try {
                IDTSVirtualInputColumnCollection90 virtualInputColumnCollection = this.VirtualInput.VirtualInputColumnCollection;
                int virtualInputColumnsCount = virtualInputColumnCollection.Count;

                args.AvailableColumns = new AvailableColumnElement[virtualInputColumnsCount];
                for (int i = 0; i < virtualInputColumnsCount; i++) {
                    IDTSVirtualInputColumn90 virtualInputColumn = virtualInputColumnCollection[i];
                    args.AvailableColumns[i].Selected = (virtualInputColumn.UsageType != DTSUsageType.UT_IGNORED);
                    args.AvailableColumns[i].AvailableColumn = new DataFlowElement (virtualInputColumn.Name, virtualInputColumn);
                }
            } catch (Exception ex) {
                this.ReportErrors (ex);
            }
        }


        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        void form_SetInputOutputColumns (object sender, SetInputOutputColumnsArgs args) {
            Debug.Assert (args.VirtualColumn != null, "Invalid arguments passed from the UI");

            this.ClearErrors ();
            try {
                IDTSInput90 input = this.ComponentMetadata.InputCollection[0];

                IDTSVirtualInputColumn90 virtualInputColumn = args.VirtualColumn.Tag as IDTSVirtualInputColumn90;
                if (virtualInputColumn == null) {
                    throw new ApplicationException (Properties.Resources.UIisInconsistentState);
                }

                int lineageId = virtualInputColumn.LineageID;

                IDTSOutput90 distinctOutput = this.ComponentMetadata.OutputCollection[0]; 

                IDTSInputColumn90 inputColumn = this.DesigntimeComponent.SetUsageType (input.ID, this.VirtualInput, lineageId, DTSUsageType.UT_READONLY);

                args.GeneratedColumns.InputColumn = new DataFlowElement (inputColumn.Name, inputColumn);
                args.GeneratedColumns.DistinctColumn = FindMappedOutputColumnElement (lineageId, distinctOutput);
            } catch (Exception ex) {
                this.ReportErrors (ex);
                args.CancelAction = true;
            }
        }

        /// <summary>
        /// 
        /// This method is called by the form when a columns is to be deleted from the used ones
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="args"></param>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        void form_DeleteInputOutputColumns (object sender, SetInputOutputColumnsArgs args) {
            Debug.Assert (args.VirtualColumn != null, "Invalid arguments passed from the UI");

            this.ClearErrors ();
            try {
                IDTSInput90 input = this.ComponentMetadata.InputCollection[0];

                IDTSVirtualInputColumn90 virtualInputColumn = args.VirtualColumn.Tag as IDTSVirtualInputColumn90;
                if (virtualInputColumn == null) {
                    throw new ApplicationException (Properties.Resources.UIisInconsistentState);
                }

                int lineageId = virtualInputColumn.LineageID;

                this.DesigntimeComponent.SetUsageType (input.ID, this.VirtualInput, lineageId, DTSUsageType.UT_IGNORED);
            } catch (Exception ex) {
                this.ReportErrors (ex);
                args.CancelAction = true;
            }
        }

        /// <summary>
        /// 
        /// Gets all output columns from a given output indexed by lineage IDs of
        /// mapped inputs columns.
        /// 
        /// </summary>
        /// <param name="output">passed output object</param>
        /// <returns>generated hash table</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes")]
        private static Dictionary<int, IDTSOutputColumn90> GetOutputColumnsDictionary (IDTSOutput90 output) {
            Debug.Assert (output != null);

            IDTSOutputColumnCollection90 outputColumnCollection = output.OutputColumnCollection;
            int outputColumnCount = outputColumnCollection.Count;

            Dictionary<int, IDTSOutputColumn90> outputColumnByLineageID = new Dictionary<int, IDTSOutputColumn90> (outputColumnCount);

            foreach (IDTSOutputColumn90 outputColumn in outputColumnCollection) {
                IDTSCustomProperty90 property = GetCustomProperty (InputColumnLineageIdPropName, outputColumn.CustomPropertyCollection);
                if (property != null && property.Value is int) {
                    int lineageId = (int) property.Value;
                    if (outputColumnByLineageID.ContainsKey (lineageId)) {
                        //
                        // This should not happen. There should not be more than one output column,
                        // in the same collection, pointing to the same input column.
                        //
                        throw new ApplicationException ("Duplicate output column mapping found.");
                    } else {
                        outputColumnByLineageID.Add (lineageId, outputColumn);
                    }
                }
            }

            return outputColumnByLineageID;
        }

        /// <summary>
        /// Get output column from the given dictionary mapped to the given lineage ID.
        /// </summary>
        /// <param name="lineageID"></param>
        /// <param name="outputColumnsByLineageID"></param>
        /// <returns></returns>
        private static DataFlowElement GetMappedOutputColumnElement (int lineageID, Dictionary<int, IDTSOutputColumn90> outputColumnsByLineageID) {
            Debug.Assert (outputColumnsByLineageID != null);

            if (outputColumnsByLineageID.ContainsKey (lineageID)) {
                IDTSOutputColumn90 outputColumn = outputColumnsByLineageID[lineageID];
                return new DataFlowElement (outputColumn.Name, outputColumn);
            } else {
                return new DataFlowElement ();
            }
        }

        /// <summary>
        /// 
        /// Look for output column in the given output collection that is mapped using the given lineage ID.
        /// 
        /// </summary>
        /// <param name="lineageID"></param>
        /// <param name="output"></param>
        /// <returns>data flow element that wraps output column</returns>
        private static DataFlowElement FindMappedOutputColumnElement (int targetLineageID, IDTSOutput90 output) {
            Debug.Assert (output != null);

            IDTSOutputColumnCollection90 outputColumnCollection = output.OutputColumnCollection;

            DataFlowElement retOutputElement = new DataFlowElement ();

            foreach (IDTSOutputColumn90 outputColumn in outputColumnCollection) {
                IDTSCustomProperty90 property = GetCustomProperty (InputColumnLineageIdPropName, outputColumn.CustomPropertyCollection);
                if (property != null && property.Value is int) {
                    int lineageId = (int) property.Value;
                    if (lineageId == targetLineageID) {
                        retOutputElement = new DataFlowElement (outputColumn.Name, outputColumn);
                        break;
                    }
                }
            }

            return retOutputElement;
        }

    }
}
